#include <stdio.h>
#include <stdlib.h>
#include "emulator.h"

short etching_zone = 0, galvanizing_zone = 0;

char zone_is_busy(short zone) {
    for (short i = 0; i < 10; i++) {
        if (barrels[i].flags.is_exist && barrels[i].zone == zone) {
            return 1;
        }
    }
    return 0;
}

static short get_robot_barrel(char robot_id) {
    for (short i = 0; i < BARRELS_COUNT; i++) {
        if (barrels[i].flags.robot == robot_id) {
            return i;
        }
    }
    return -1;
}

// TODO обновить метод
// вернет можно ли ехать и главное куда ехать, если можно (нельзя если вернулось значение < 0)
// -1 вернет что перемещать нельзя
// -2 вернет если требуется атомарная операция пассивации
short can_move(struct barrel* bar) {
    // сразу отсекаем варианты, при которых невозможно переместить барабан
    if (!bar->flags.is_exist) {
        return -1;
    }

    if (bar->software_timer > 0) {
        return -1;
    }

    if (bar->flags.robot != 0) {
        return -2;
    }

    if (!schedulerOneRobotMode) {
        printf("WARMING: нет проверки того, что для перемещения барабана не мешает второй робот\n");
    }

    // дальше нужно проверить, что можно передвигать бочку

    switch (bar->curr_process) {
        case PROCESS_NONE:
            // загрузка, нужно обезжиривание (зона 2)
            if (!zone_is_busy(2)) {
                return 2;
            }
            break;

        case PROCESS_DEFATTING:
            // обезжиривание, нужна промывка 1А (зона 3)
            if (!zone_is_busy(3)) {
                return 3;
            }
            break;

        case PROCESS_WASHING_1A:
            // промывка 1А, нужна промывка 1Б (зона 4)
            if (!zone_is_busy(4)) {
                return 4;
            }
            break;

        case PROCESS_WASHING_1B:
            // промывка 1Б, нужно травление (зоны 5-6)
            if (!zone_is_busy(5 + etching_zone)) {
                return 5 + etching_zone;
            }
            break;

        case PROCESS_ETCHING:
            // травление, нужна промывка 2А (зона 7)
            if (!zone_is_busy(7)) {
                return 7;
            }
            break;

        case PROCESS_WASHING_2A:
            // промывка 2А, нужна промывка 2Б (зона 8)
            if (!zone_is_busy(8)) {
                return 8;
            }
            break;

        case PROCESS_WASHING_2B:
            // промывка 2Б, нужно цинкование (зоны 9-16)
            if (!zone_is_busy(9 + galvanizing_zone)) {
                return 9 + galvanizing_zone;
            }
            break;

        case PROCESS_GALVANIZING:
            // цинкование, требуется чтобы в зонах 17-22 было максимум 2 барабана (3 барабана для этой части линии - максимум)
            {
                short count = 0;
                for (short i = 17; i <= 22; i++) {
                    if (zone_is_busy(i)) {
                        count++;
                    }
                }
                if (count < 3 && !zone_is_busy(17)) {
                    return 17;
                }
            }
            break;

        case PROCESS_WASHING_3A:
            // промывка 3А, перекладываем в промывку 3Б (зона 18)
            if (!zone_is_busy(18)) {
                return 18;
            }
            break;

        case PROCESS_WASHING_3B:
            // это перед пассивацией, требует свободную промывку 4А (зона 20) и на всякий случай свободную пассивацию (зона 19)
            if (!zone_is_busy(19) && !zone_is_busy(20)) {
                return -2;
            }
            // это атомарная операция, по идее вносить барабан в пассивацию нельзя
            break;
        case PROCESS_PASSIVATION:
            // процесс пассивации, нужна промывка 4B (зона 21) (потому что сейчас я в 4A)
            if (!zone_is_busy(21)) {
                return 21;
            }
            break;
        case PROCESS_WASHING_4B:
            // процесс пассивации, нужна промывка 4B (зона 21) (потому что сейчас я в 4A)
            if (!zone_is_busy(22)) {
                return 22;
            }
            break;
        case PROCESS_RETURN_1:
            // последняя промывка, нужно разрешение на выгрузку
            if (schedulerUnloadButton) {
                // нужно промывку 3б (зона 10) и загрузку 0
                if (!zone_is_busy(17) || !zone_is_busy(0)) {
                    return 17;
                }
            }
            break;
        case PROCESS_RETURN_2:
            if (schedulerOneRobotMode) {
                if (!schedulerUnloadButton) {
                    break;
                }
            }

            // нужна свободная выгрузка
            if (!zone_is_busy(0)) {
                return 0;
            }

            break;
    }
    return -1;
}


struct scheduler_task {
    short start_zone; // стартовая зона
    short dest_zone; // конечная зона
    short priority; // приоритет, чем больше тем выше, по умолчанию 0
};

// выставляет приоритет операции
short get_operation_priority(short barrel_id) {
    // сделать приоритет на барабан, который больше всего ждет

//    task->priority = 0;
//    switch (barrel_process) {
//        case PROCESS_RETURN_2:
//            break;
//    }

    return 0;
}


struct barrel makeBarrel(short flags, short zone, short timer, short process) {
    struct barrel b;
    b.flags.raw_word = flags;
    b.zone = zone;
    b.software_timer = timer;
    b.curr_process = process;
    b.time_defatting = 12;
    b.time_washing_1a = 4;
    b.time_washing_1b = 6;
    b.time_etching = 16;
    b.time_washing_2a = 6;
    b.time_washing_2b = 8;
    b.time_galvanizing = 30;
    b.time_washing_3a = 8;
    b.time_washing_3b = 10;
    b.time_passivation = 3;
    b.time_washing_4a = 12;
    b.time_washing_4b = 14;
    return b;
}


static short scheduler_find_task(struct scheduler_task* tasks, const short curr_pos) {
    // ищем первый барабан слева, и ближайший справа
    short left = -1, right = -1;
    for (short i = 0; i < BARRELS_COUNT; i++) {
        short target = tasks[i].start_zone; // фактическая зона откуда тащить барабан
        if (tasks[i].dest_zone == -2) {
            target = 18;
        } else if (tasks[i].dest_zone < 0) {
            continue;
        }

        if (curr_pos <= target) {
            // это таск справа, надо найти ближайший
            if (right == -1) {
                right = i;
            } else {
                if (barrels[right].zone > target) {
                    right = i;
                }
            }
        } else {
            // таск слева, ищем максимально дальний (с минимальной зоной)
            if (left == -1) {
                left = i;
            } else {
                if (barrels[left].zone > target) {
                    left = i;
                }
            }
        }
    }

    // итого есть результат: есть ли таски, которые надо тащить вперед (и если надо то какой ближний), и есть первый таск
    if (left < 0) {
        return right; // вернем таск справа (если его нет, в переменной будет -1)
    }
    if (right < 0) {
        return left; // если вдруг задачи справа не оказалось, вернем задачу слева если есть
    }

    // вычисляем что ближе
    short ld = curr_pos - left; // левая дельта
    short rd = right - curr_pos; // правая дельта

    // дальше сравниваем дельты
    // по идее если они равны то с бОльшим приоритетом робот поедет в левую часть
    // а левую дельту вообще уменьшу на 1, чтобы цель слева казалась ближе
    if (rd > ld - 1) {
        return left;
    } else {
        return right;
    }
}


void schedule_robot_1() {
    static short transaction_state = 0;

    // начало транзакции

    if (transaction_state == 0) {
        // формируем список задач
        struct scheduler_task tasks[BARRELS_COUNT];
        for (short i = 0; i < BARRELS_COUNT; i++) {
            // для каждой задачи:
            tasks[i].start_zone = barrels[i].zone;
            // определяем можно ли ее выполнить и что вообще нужно выполнить
            tasks[i].dest_zone = can_move(barrels + i);
        }

        // найти подходящую задачу
        if (schedulerOneRobotMode) {
            short target_task = scheduler_find_task(tasks, robot1.dx.current_zone);

            if (target_task >= 0) {
                // создаем транзакцию
                if (tasks[target_task].dest_zone == -2) {
                    // пассивация
                    printf("Get work: passivate barrel %d\n", target_task);
                    robot1_cmd.cmd = 3; // пассивация
                } else {
                    printf("Get work: move barrel %d from %d to %d\n",
                           target_task, tasks[target_task].start_zone, tasks[target_task].dest_zone);
                    robot1_cmd.cmd = 2; // везем барабан
                    robot1_cmd.args[0] = tasks[target_task].start_zone;
                    robot1_cmd.args[1] = tasks[target_task].dest_zone;
                }
//            barrels[target_barrel].zone = target_zone;
                barrels[target_task].flags.robot = 1;

                // TODO сделать нормальное переключение зон с учетом тех, что можно отключить
                // TODO добавить ограничение - нельзя отключить сразу все зоны цинкования или травления
                if (tasks[target_task].dest_zone == 3 || tasks[target_task].dest_zone == 4) {
                    etching_zone ^= 0x1; // переключаем следующую зону
                }

                if (tasks[target_task].dest_zone >= 9 && tasks[target_task].dest_zone <= 17) {
                    galvanizing_zone = (galvanizing_zone + 1) % 8;
                }

                if (tasks[target_task].start_zone == 22) {
                    // выгрузка, снимаем кнопку выгрузки
                    schedulerUnloadButton = 0;
                }

                transaction_state = 1;
            }
        } else {
            printf("WARMING: support only one robot mode\n");
        }
    } else if (transaction_state == 1) {
        if (robot1_cmd.cmd == 0) {
            transaction_state = 2;
        }
    } else {
        // post transaction
        short barrel = get_robot_barrel(1);

        struct barrel* b = &barrels[barrel];

        if (barrel >= 0) {
            b->flags.is_up = 0;
            b->flags.robot = 0;
            if (robot1_cmd.cmd == 2) {
                b->zone = robot1_cmd.args[2];
            } else if (robot1_cmd.cmd == 3) {
                b->zone = 20;
            }
        }

        switch (b->curr_process) {
            // case curr: bar->curr_process = next; bar->software_timer = bar->time; break

            // после загрузки: ставим обезжиривание
            case PROCESS_NONE: b->curr_process = PROCESS_DEFATTING; b->software_timer = b->time_defatting; break;

            // после обезжира: ставим промывку 1а
            case PROCESS_DEFATTING: b->curr_process = PROCESS_WASHING_1A; b->software_timer = b->time_washing_1a; break;

            // после промывки 1а: ставим промывку 1б
            case PROCESS_WASHING_1A: b->curr_process = PROCESS_WASHING_1B; b->software_timer = b->time_washing_1b; break;

            // после промывки 1а: ставим травление
            case PROCESS_WASHING_1B: b->curr_process = PROCESS_ETCHING; b->software_timer = b->time_etching; break;

            // после травления: ставим промывку 2а
            case PROCESS_ETCHING: b->curr_process = PROCESS_WASHING_2A; b->software_timer = b->time_washing_2a; break;

            // после промывки 2а: ставим промывку 2б
            case PROCESS_WASHING_2A: b->curr_process = PROCESS_WASHING_2B; b->software_timer = b->time_washing_2b; break;

            // после промывки 2б: ставим цинкование
            case PROCESS_WASHING_2B: b->curr_process = PROCESS_GALVANIZING; b->software_timer = b->time_galvanizing; break;

            // после цинкования: ставим промывку 3а
            case PROCESS_GALVANIZING: b->curr_process = PROCESS_WASHING_3A; b->software_timer = b->time_washing_3a; break;

            // после промывки 3а: ставим промывку 1б
            case PROCESS_WASHING_3A: b->curr_process = PROCESS_WASHING_3B; b->software_timer = b->time_washing_3b; break;

            // после промывки 3б: ставим пассивацию
            case PROCESS_WASHING_3B: b->curr_process = PROCESS_PASSIVATION; break;

            // после пассивации, барабан в промывке 4а: ставим промывку 4б
            case PROCESS_PASSIVATION: b->curr_process = PROCESS_WASHING_4B; b->software_timer = b->time_washing_4b; break;

            // после промывки 4б: дальше выгрузка
            case PROCESS_WASHING_4B:
                if (schedulerOneRobotMode) {
                    b->curr_process = PROCESS_RETURN_2;
                } else {
                    b->curr_process = PROCESS_RETURN_1;
                }
                b->software_timer = 0;
                break;

            case PROCESS_RETURN_1: b->curr_process = PROCESS_RETURN_2; b->software_timer = 0; break;

            case PROCESS_RETURN_2:
            default:
                b->flags.is_exist = 0;
        }

        transaction_state = 0;
    }
}

// вернет false если удалось вставить барабан, иначе true
char create_barrel_in_load(short zone) {
    for (int i = 0; i < BARRELS_COUNT; i++) {
        if (!barrels[i].flags.is_exist) {
            barrels[i] = makeBarrel(1, zone, 0, PROCESS_NONE);
            return 0;
        }
    }
    return 1;
}

/*
=== ЦИКЛОГРАММА ПЕРЕТАСКИВАНИЯ БАРАБАНА ===
// NOTE первой командой на любую транзакцию должна стоять команда опустить траверсу (в 22 зоне мы никогда не закончим, за нее не беспокоится)
* опустить траверсу
если зона изъятия != промывка 3б
	если текущая зона != зона иъятия
		если зона изъятия == 22
			* встать на смещенную
			* поднять траверсу
			* съебать в 21 зону
			* встать на смещенную
			* опустить траверсу не до конца
			* съебать в 22 зону
		иначе
			* съебать в зону изъятия
	* поднять траверсу
	если зона изъятия != 22 и зона изъятия != 0:
		* ждать скапывания (зависит от зоны)
	* ехать в зону назначения
	если зона назначения == 22
		* опустить траверсу не до конца
		* съебать в 21 зону
		если ЭНКОДЕРЫ СТАРЫЕ (по умолчанию)
			* поднять траверсу
        * опустить траверсу
	иначе
		* опустить траверсу
		если зона назначения != 1
            * установить время ожидания барабана (зависит от зоны)
иначе
	если текущая зона != промывка 3б
		* съебать в промывку 3б
	* поднять траверсу
	* съебать в пассивацию
	* опустить траверсу
	* поставить время ожидания барабана в <время пассивации>
	* поднять траверсу
	* съебать в зону промывка 4а
	* опустить траверсу
	* установить время ожидания барабана (для промывки 4а)
 */

void scheduler_main() {
    // тут должно быть удаление барабана из зоны 1, если он там есть
//    if (schedulerLoadButton1) {
//        schedulerLoadButton1 = create_barrel_in_load(0);
//    }

    // тут возможна только вставка барабанов
    if (schedulerLoadButton2) {
        schedulerLoadButton2 = create_barrel_in_load(1);
    }

    if (schedulerSoftwareTimer) {
        schedulerSoftwareTimer = 0;
        for (int i = 0; i < 10; i++) {
            if (barrels[i].software_timer > -9999) {
                barrels[i].software_timer--;
            }
        }
    }
    if (robot1_cmd.cmd == 0) {
        schedule_robot_1();
    }

    // пока без второго робота
//    if (robot2_cmd.cmd == 0) {
//        robot2_cmd.cmd = 2; // пиздуем в зону
//        robot2_cmd.args[0] = rand() % 23;
//    }
}
